"""Server wrapper."""

from argparse import ArgumentParser, Namespace
from logging import INFO, WARNING, basicConfig, getLogger
from pathlib import Path
from subprocess import Popen
from time import sleep

from dzdsu.constants import JSON_FILE, XVFB
from dzdsu.server import load_servers


__all__ = ['main']


LOGGER = getLogger('dzdsw')


def get_args(description: str = __doc__) -> Namespace:
    """Return the parsed command line arguments."""

    parser = ArgumentParser(description=description)
    parser.add_argument(
        'server', default=Path.cwd().name,
        help='the name of the server to start'
    )
    parser.add_argument(
        '-f', '--servers-file', type=Path, default=JSON_FILE, metavar='file',
        help='servers JSON file path'
    )
    parser.add_argument(
        '-v', '--verbose', action='store_true', help='verbose logging'
    )
    return parser.parse_args()


def main() -> int:
    """Starts the DayZ server."""

    args = get_args()
    basicConfig(level=INFO if args.verbose else WARNING)
    servers = load_servers(args.servers_file)
    env = None

    try:
        server = servers[args.server]
    except KeyError:
        LOGGER.error('No such server: %s', args.server)
        return 2

    if server.wine:
        Popen(XVFB, cwd=server.base_dir)
        env = {'DISPLAY': ':0'}
        sleep(3)    # Give Xvfb some time to start up.

    Popen(server.command, cwd=server.base_dir, env=env)
    return 0



"""Common constants."""

from os import environ, name
from pathlib import Path


__all__ = [
    'BATTLEYE_GLOB',
    'CONFIG_FILE',
    'DAYZ_APP_ID',
    'DAYZ_SERVER_APP_ID',
    'ITALIC',
    'JSON_FILE',
    'LINK',
    'MODS_BASE_DIR',
    'MODS_DIR',
    'SERVER_EXECUTABLE',
    'SERVER_EXECUTABLE_WIN',
    'STEAMCMD',
    'STEAMCMD_WINE',
    'WINE',
    'WORKSHOP_URL',
    'XVFB'
]


BATTLEYE_GLOB = 'battleye/beserver_x64*.cfg'
CONFIG_FILE = 'serverDZ.cfg'
DAYZ_APP_ID = 221100
DAYZ_SERVER_APP_ID = 223350
MODS_BASE_DIR = Path('steamapps/workshop/content/')
MODS_DIR = MODS_BASE_DIR / str(DAYZ_APP_ID)
SERVER_EXECUTABLE_WIN = 'DayZServer_x64.exe'
STEAMCMD = 'steamcmd'
STEAMCMD_WINE = '.steamcmd/steamcmd.exe'
WINE = '/usr/bin/wine'
WORKSHOP_URL = 'https://steamcommunity.com/sharedfiles/filedetails/?id={}'
XVFB = ('/usr/bin/Xvfb', ':0', '-screen', '0', '1024x768x16')

ITALIC = '\033[3m{}\033[0m'
LINK = '\x1b]8;;{url}\x1b\\{text}\x1b]8;;\x1b\\'

if name == 'nt':
    JSON_FILE = Path(environ.get('PROGRAMFILES')) / 'dzsrv' / 'servers.json'
    SERVER_EXECUTABLE = SERVER_EXECUTABLE_WIN
elif name == 'posix':
    JSON_FILE = Path('/etc/dzservers.json')
    SERVER_EXECUTABLE = 'DayZServer'
else:
    raise OSError('Unsupported operating system.')



"""Server utility."""

from argparse import ArgumentParser, Namespace
from logging import INFO, WARNING, basicConfig, getLogger
from pathlib import Path

from dzdsu.constants import JSON_FILE
from dzdsu.mods import print_mods
from dzdsu.server import Server, load_servers
from dzdsu.update import Updater


__all__ = ['main']


LOGGER = getLogger('dzdsu')


def get_args(description: str = __doc__) -> Namespace:
    """Return the parsed command line arguments."""

    parser = ArgumentParser(description=description)
    parser.add_argument(
        'server', default=Path.cwd().name,
        help='the name of the server to start'
    )
    parser.add_argument(
        '-f', '--servers-file', type=Path, default=JSON_FILE, metavar='file',
        help='servers JSON file path'
    )
    parser.add_argument(
        '-C', '--clean-mods', action='store_true', help='remove unused mods'
    )
    parser.add_argument(
        '-F', '--fix-paths', action='store_true', help='fix mod file paths'
    )
    parser.add_argument(
        '-I', '--installed-mods', action='store_true',
        help='list installed mods'
    )
    parser.add_argument(
        '-K', '--install-keys', action='store_true', help='install mod keys'
    )
    parser.add_argument(
        '-M', '--list-mods', action='store_true', help="list the server's mods"
    )
    parser.add_argument(
        '-S', '--list-server-mods', action='store_true',
        help="list the server's server mods"
    )
    parser.add_argument(
        '-U', '--update', metavar='steam_user',
        help='update server and/or mods'
    )
    parser.add_argument(
        '-m', '--update-mods', action='store_true',
        help="update the server's mods"
    )
    parser.add_argument(
        '-s', '--update-server', action='store_true', help='update server'
    )
    parser.add_argument(
        '--overwrite', action='store_true', help="overwrite existing key files"
    )
    parser.add_argument(
        '-v', '--verbose', action='store_true', help='verbose logging output'
    )
    return parser.parse_args()


def clean_mods(server: Server) -> None:
    """Remove unused mods."""

    for installed_mod in server.unused_mods:
        LOGGER.info('Removing unused mod: %s', installed_mod.mod)
        installed_mod.remove()


def install_keys(server: Server) -> None:
    """Installs the keys for all mods of the server."""

    for installed_mod in server.installed_mods:
        for key in installed_mod.bikeys:
            if (installed := server.base_dir / 'keys' / key.name).exists():
                LOGGER.info('Key "%s" already installed.', key.name)
                continue

            with key.open('rb') as src, installed.open('wb') as dst:
                dst.write(src.read())


def fix_mod_paths(server: Server) -> None:
    """Fix paths of the server mods."""

    for installed_mod in server.installed_mods:
        installed_mod.fix_paths()


def update(server: Server, args: Namespace) -> None:
    """Perform server and mod updates."""

    updater = Updater(server, args.update)

    if args.update_server:
        updater.update_server()

    if args.update_mods:
        updater.update_mods()

    updater()


def main() -> int:
    """Update mods."""

    args = get_args()
    basicConfig(level=INFO if args.verbose else WARNING)
    servers = load_servers(args.servers_file)

    try:
        server = servers[args.server]
    except KeyError:
        LOGGER.error('No such server: %s', args.server)
        return 2

    if args.clean_mods:
        clean_mods(server)

    if args.update:
        update(server, args)

    if args.fix_paths:
        fix_mod_paths(server)

    if args.install_keys:
        install_keys(server)

    if args.list_mods:
        print_mods(server.mods)

    if args.list_server_mods:
        print_mods(server.server_mods)

    if args.installed_mods:
        print_mods(sorted(map(
            lambda installed_mod: installed_mod.mod,
            server.installed_mods
        )))

    return 0



"""Server representation."""

from configparser import SectionProxy
from itertools import chain
from json import load
from pathlib import Path
from typing import Any, Iterator, NamedTuple

from dzdsu.constants import BATTLEYE_GLOB
from dzdsu.constants import DAYZ_SERVER_APP_ID
from dzdsu.constants import JSON_FILE
from dzdsu.constants import MODS_DIR
from dzdsu.constants import SERVER_EXECUTABLE
from dzdsu.constants import SERVER_EXECUTABLE_WIN
from dzdsu.constants import WINE
from dzdsu.mods import Mod, ModMetadata, InstalledMod, mods_str
from dzdsu.params import ServerParams
from dzdsu.parsers import parse_battleye_cfg, parse_server_cfg, wine_path


__all__ = ['Server', 'load_servers']


class Server(NamedTuple):
    """A server."""

    name: str
    app_id: int
    base_dir: Path
    mods: list[Mod]
    server_mods: list[Mod]
    params: ServerParams
    wine: bool

    @classmethod
    def from_json(cls, name: str, json: dict):
        """Creates a Server instance from a JSON-ish dict."""
        return cls(
            name,
            json.get('appId', DAYZ_SERVER_APP_ID),
            Path(json['basedir']),
            [Mod.from_value(mod) for mod in (json.get('mods') or [])],
            [Mod.from_value(mod) for mod in (json.get('serverMods') or [])],
            ServerParams.from_json(json.get('params') or {}),
            json.get('wine', False)
        )

    @property
    def executable_args(self) -> Iterator[str]:
        """Yields arguments for the server executable."""
        yield from self.params.executable_args

        if mods := mods_str(mod for mod in self.mods if mod.enabled):
            yield f'-mod={mods}'

        if mods := mods_str(mod for mod in self.server_mods if mod.enabled):
            yield f'-serverMod={mods}'

    @property
    def executable(self) -> list[str]:
        """Returns the executable args."""
        if self.wine:
            return [WINE, str(self.base_dir / SERVER_EXECUTABLE_WIN)]

        return [str(self.base_dir / SERVER_EXECUTABLE)]

    @property
    def command(self) -> list[str]:
        """Returns the full command for running the server."""
        return [*self.executable, *self.executable_args]

    @property
    def install_dir(self) -> str:
        """Returns the installation directory for steamcmd."""
        if self.wine:
            return wine_path(self.base_dir)

        return str(self.base_dir)

    @property
    def mods_dir(self) -> Path:
        """Returns the server's mods directory."""
        return self.base_dir / MODS_DIR

    @property
    def battleye_cfg_file(self) -> Path:
        """Returns the BattlEye RCon config file."""
        for path in self.base_dir.glob(BATTLEYE_GLOB):
            if path.is_file():
                return path

        raise FileNotFoundError(BATTLEYE_GLOB)

    @property
    def battleye_cfg(self) -> dict[str, str | int | bool]:
        """Returns the BattlEye RCon configuration."""
        with self.battleye_cfg_file.open('r', encoding='utf-8') as file:
            return dict(parse_battleye_cfg(file))

    @property
    def installed_mods_metadata(self) -> Iterator[ModMetadata]:
        """Yields metadata of installed mods."""
        for filename in self.mods_dir.glob('*/meta.cpp'):
            yield ModMetadata.from_file(filename)

    @property
    def config_file(self) -> Path:
        """Returns the config file path."""
        return self.base_dir / 'serverDZ.cfg'

    @property
    def config(self) -> SectionProxy:
        """Returns the configuration settings."""
        with self.config_file.open('r', encoding='utf-8') as file:
            return parse_server_cfg(file)

    @property
    def installed_mods(self) -> Iterator[InstalledMod]:
        """Yields installed mods."""
        for meta in self.installed_mods_metadata:
            yield InstalledMod(meta.publishedid, self.base_dir)

    @property
    def used_mods(self) -> Iterator[Mod]:
        """Yields used mods."""
        return chain(self.mods, self.server_mods)

    @property
    def unused_mods(self) -> Iterator[InstalledMod]:
        """Yields unused mods."""
        used_ids = {mod.id for mod in self.used_mods}

        for meta in self.installed_mods_metadata:
            if meta.publishedid not in used_ids:
                yield InstalledMod(meta.publishedid, self.base_dir)


def load_servers_json(file: Path) -> dict[str, Any]:
    """Loads servers from a JSON file."""

    with file.open('r') as json:
        return load(json)


def load_servers(file: Path = JSON_FILE) -> dict[str, Server]:
    """Loads servers."""

    return {
        name: Server.from_json(name, json)
        for name, json in load_servers_json(file).items()
    }



"""Modifications from the Steam workshop."""

from __future__ import annotations
from pathlib import Path
from shutil import rmtree
from typing import Iterable, Iterator, NamedTuple, Optional

from dzdsu.constants import ITALIC
from dzdsu.constants import LINK
from dzdsu.constants import MODS_DIR
from dzdsu.constants import WORKSHOP_URL


__all__ = ['Mod', 'ModMetadata', 'InstalledMod', 'mods_str', 'print_mods']


class Mod(NamedTuple):
    """A server mod."""

    id: int
    name: Optional[str] = None
    enabled: bool = True

    def __str__(self) -> str:
        return LINK.format(url=self.url, text=self.name or self.id)

    @classmethod
    def from_id(cls, ident: int, *, name: Optional[str] = None) -> Mod:
        """Creates a mod from an ID."""
        if ident == 0:
            raise ValueError(f'Invalid mod ID: {ident}')

        if ident < 0:
            return cls(abs(ident), name, enabled=False)

        return cls(ident, name)

    @classmethod
    def from_json(cls, json: dict[str, int | str]) -> Mod:
        """Creates a mod from a JSON-ish dict."""
        return cls.from_id(json['id'], name=json.get('name'))

    @classmethod
    def from_value(cls, value: int | dict[str, int | str]) -> Mod:
        """Creates a mod from an int or JSON value."""
        if isinstance(value, int):
            return cls.from_id(value)

        if isinstance(value, dict):
            return cls.from_json(value)

        raise TypeError(f'Cannot create mod from: {value} ({type(value)})')

    @property
    def path(self) -> Path:
        """Returns the relative path to the local mod directory."""
        return MODS_DIR / str(self.id)

    @property
    def url(self) -> str:
        """Returns the Steam Workshop URL."""
        return WORKSHOP_URL.format(self.id)


class ModMetadata(NamedTuple):
    """Metadata of a mod."""

    protocol: int
    publishedid: int
    name: str
    timestamp: int

    @classmethod
    def from_dict(cls, dct: dict) -> ModMetadata:
        """Creates mod metadata from the given dict."""
        return cls(
            int(dct['protocol']),
            int(dct['publishedid']),
            dct['name'].strip('"'),
            int(dct['timestamp'])
        )

    @classmethod
    def from_lines(cls, lines: Iterable[str]) -> ModMetadata:
        """Creates mod metadata from the given lines."""
        return cls.from_dict({
            key: value.rstrip(';') for key, value in (
                line.split(' = ') for line in (
                    line.strip() for line in lines
                ) if line
            )
        })

    @classmethod
    def from_file(cls, filename: Path) -> ModMetadata:
        """Reads the mod metadata from the given file."""
        with filename.open('r', encoding='utf-8') as file:
            return cls.from_lines(file)


class InstalledMod(NamedTuple):
    """Represents an installed mod."""

    id: int
    base_dir: Path

    @property
    def mod(self) -> Mod:
        """Returns a Mod object."""
        return Mod(self.id, self.metadata.name)

    @property
    def path(self) -> Path:
        """Returns the relative path to the local mod directory."""
        return self.base_dir / MODS_DIR / str(self.id)

    @property
    def addons(self) -> Path:
        """Returns the path to the addons directory."""
        return self.path / 'addons'

    @property
    def keys(self) -> Path:
        """Returns the path to the keys directory."""
        return self.path / 'keys'

    @property
    def metadata_file(self) -> Path:
        """Returns the path to the metadata file."""
        return self.path / 'meta.cpp'

    @property
    def metadata(self) -> ModMetadata:
        """Returns the mod metadata."""
        return ModMetadata.from_file(self.metadata_file)

    @property
    def pbos(self) -> Iterator[Path]:
        """Yields paths to the .pbo files."""
        return self.addons.glob('*.pbo')

    @property
    def bikeys(self) -> Iterator[Path]:
        """Yields paths to the *.bikey files."""
        return self.keys.glob('*.bikey')

    def fix_paths(self) -> None:
        """Links paths to lower-case."""
        if (addons := self.path / 'Addons').is_dir():
            link_to_lowercase(addons)

        if (keys := self.path / 'Keys').is_dir():
            link_to_lowercase(keys)

        for pbo in self.pbos:
            link_to_lowercase(pbo)

    def remove(self) -> None:
        """Removes this mod."""
        rmtree(self.path)


def link_to_lowercase(path: Path) -> None:
    """Creates a symlink with the path names in lower case."""

    if (filename := path.name) == (lower := filename.lower()):
        return

    if (symlink := path.parent / lower).exists():
        return

    symlink.symlink_to(filename)


def mods_str(mods: Iterable[Mod], sep: str = ';') -> str:
    """Returns a string representation of the given mods."""

    return sep.join(str(mod.path) for mod in mods)


def print_mods(mods: Iterable[Mod]) -> None:
    """Lists the respective mods."""

    for mod in mods:
        print(mod if mod.enabled else ITALIC.format(mod))



"""Config file parsers."""

from configparser import ConfigParser, SectionProxy
from pathlib import Path
from re import fullmatch
from typing import Iterable, Iterator


__all__ = ['parse_battleye_cfg', 'parse_server_cfg', 'wine_path']


def parse_battleye_value(key: str, value: str) -> bool | int | str:
    """Parses a BattlEye config value."""

    if key == 'RConPort':
        return int(value)

    if key == 'RestrictRCon':
        return bool(int(value))

    return value


def parse_battleye_cfg(
        lines: Iterable[str]
) -> Iterator[tuple[str, bool | int | str]]:
    """Yields key / value pairs of the given config file."""

    for line in lines:
        if not (line := line.strip()) or line.startswith('#'):
            continue

        key, value = line.split(maxsplit=1)
        yield key, parse_battleye_value(key, value)


def server_cfg_to_ini(lines: Iterable[str]) -> Iterator[tuple[str, str]]:
    """Yields lines of an INI-style representation of the server config."""

    for line in lines:
        if not (line := line.strip()):
            continue

        if not (match := fullmatch(r'^(\w+)\s*=\s*(.+);.*', line)):
            continue

        key, value = match.groups()
        yield key, value.strip('"')


def parse_server_cfg(
        lines: Iterable[str], *, section: str = 'Server'
) -> SectionProxy:
    """Yields key / value pairs of the given server config."""

    config = ConfigParser()
    config.read_dict({section: dict(server_cfg_to_ini(lines))})
    return config[section]


def wine_path(path: Path, *, drive: str = 'Z:') -> str:
    """Converts a path to a path for Wine."""

    if path.is_absolute():
        return '\\'.join([drive, *path.parts[1:]])

    return '\\'.join(path.parts)



"""Game and mod updates."""

from __future__ import annotations
from subprocess import CompletedProcess, run

from dzdsu.constants import DAYZ_APP_ID, STEAMCMD, STEAMCMD_WINE, WINE
from dzdsu.server import Server


__all__ = ['Updater']


class Updater:
    """SteamCMD wrapper to update server and mods."""

    def __init__(self, server: Server, steam_user_name: str):
        """Sets server name and initial command."""
        self.server = server
        self.commands = [
            '+force_install_dir', server.install_dir,
            '+login', steam_user_name
        ]

    def __str__(self):
        """Returns the command as a string."""
        return ' '.join(self.command)

    def __call__(self) -> CompletedProcess:
        """Executes the steamcmd command."""
        return run(self.command, check=True)

    @property
    def steamcmd(self) -> list[str]:
        """Returns the steamcmd command and args list."""
        if self.server.wine:
            return [WINE, str(self.server.base_dir / STEAMCMD_WINE)]

        return [STEAMCMD]

    @property
    def command(self) -> list[str]:
        """Returns the command."""
        return [*self.steamcmd, *self.commands, '+quit']

    def update_server(self) -> Updater:
        """Updates the server."""
        self.commands += ['+app_update', str(self.server.app_id), 'validate']
        return self

    def update_mods(self) -> Updater:
        """Updates the server's mods."""
        for mod in self.server.used_mods:
            self.commands += [
                '+workshop_download_item', str(DAYZ_APP_ID), str(mod.id),
                'validate'
            ]

        return self



"""Server start parameters."""

from typing import NamedTuple, Iterator, Optional

from dzdsu.constants import CONFIG_FILE


__all__ = ['ServerParams']


class ServerParams(NamedTuple):
    """Available server start parameters."""

    config: str = CONFIG_FILE
    do_logs: bool = True
    admin_log: bool = True
    net_log: bool = True
    src_allow_file_write: bool = True
    no_file_patching: bool = True
    freeze_check: bool = True
    port: Optional[int] = None
    profiles: Optional[str] = None
    cpu_count: Optional[int] = None

    @classmethod
    def from_json(cls, json: dict):
        """Creates a ServerParams instance from a JSON-ish dict."""
        return cls(
            json.get('config', CONFIG_FILE),
            json.get('doLogs', True),
            json.get('adminLog', True),
            json.get('netLog', True),
            json.get('srcAllowFileWrite', True),
            json.get('noFilePatching', True),
            json.get('freezeCheck', True),
            json.get('port'),
            json.get('profiles'),
            json.get('cpuCount')
        )

    @property
    def executable_args(self) -> Iterator[str]:
        """Yields arguments for the server executable."""
        yield f'-config={self.config}'

        if self.do_logs:
            yield '-doLogs'

        if self.admin_log:
            yield '-adminLog'

        if self.net_log:
            yield '-netLog'

        if self.src_allow_file_write:
            yield '-srcAllowFileWrite'

        if self.no_file_patching:
            yield '-noFilePatching'

        if self.freeze_check:
            yield '-freezeCheck'

        if self.port is not None:
            yield f'-port={self.port}'

        if self.profiles is not None:
            yield f'-profiles={self.profiles}'

        if self.cpu_count is not None:
            yield f'-cpuCount={self.cpu_count}'



"""DayZ dedicated server utilities."""

from dzdsu.constants import JSON_FILE
from dzdsu.mods import Mod, ModMetadata, InstalledMod, mods_str, print_mods
from dzdsu.params import ServerParams
from dzdsu.server import Server, load_servers
from dzdsu.update import Updater


__all__ = [
    'JSON_FILE',
    'InstalledMod',
    'Mod',
    'ModMetadata',
    'Server',
    'ServerParams',
    'Updater',
    'load_servers',
    'mods_str',
    'print_mods'
]



#! /usr/bin/env python
"""Installation script."""

from setuptools import setup

setup(
    name='dzdsu',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    author='Richard Neumann',
    author_email='mail@richard-neumann.de',
    python_requires='>=3.10',
    packages=['dzdsu'],
    entry_points={
        'console_scripts': [
            'dzdsu = dzdsu.utility:main',
            'dzdsw = dzdsu.wrapper:main'
        ]
    },
    url='https://github.com/conqp/dzdsu',
    license='GPLv3',
    description='DayZ dedicated server utilities',
    keywords='DayZ dedicated server utilities'
)
